#include "stdafx.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#include "uodrv.h"
#include "ftd2xx.h"
#pragma comment(lib, "ftd2xx.lib")

#pragma comment (lib,"delayimp.lib")

class queue{
	private:
		unsigned long length,size;
		PUCHAR buffer;
	public:
		queue(){
			length=0; size=128;
			buffer=new UCHAR[size];
		}
		void clear(){ length=0; }
		void enq(UCHAR d){
			if(size<=length){
				PUCHAR tmp=new UCHAR[size+128];
				memcpy(tmp,buffer,size);
				delete[] buffer;
				buffer=tmp; size+=128;
			}
			buffer[length++]=d;
		}
		unsigned long len(){return length;}
		PUCHAR buf(){return buffer;}
		~queue(){ delete[] buffer; }
};

#define SERIAL 0x80
#define READ   0x40
#define OPE    0x20	//must be 1
#define TDI    0x10
#define nCS    0x08	//must be 1
#define nCE    0x04
#define TMS    0x02
#define TCK    0x01

#define STATE0(x) {x.enq(OPE|nCS); x.enq(OPE|nCS|TCK);}
#define STATE1(x) {x.enq(OPE|nCS|TMS); x.enq(OPE|nCS|TMS|TCK);}
#define RESET(x)       {STATE1(x);STATE1(x);STATE1(x);STATE1(x);STATE1(x);}
#define RESETtoIDLE(x) {STATE0(x);}
#define IDLEtoSELDR(x) {STATE1(x);}
#define SELDRtoSIR(x)  {STATE1(x);STATE0(x);STATE0(x);}
#define SELDRtoSDR(x)  {STATE0(x);STATE0(x);}
#define PIRtoIDLE(x)   {STATE1(x);STATE1(x);STATE0(x);}
#define PDRtoIDLE(x)   PIRtoIDLE(x)
#define PIRtoSELDR(x)  {STATE1(x);STATE1(x);STATE1(x);}
#define PDRtoSELDR(x)  PIRtoSELDR(x)

#define bsf(a,b) (a)[(b)/8]|=(1<<((b)%8))
#define bcf(a,b) (a)[(b)/8]&=(0xFF^(1<<((b)%8)))

#define RCVLEN(bits) (((bits)-1)/8+((bits)-((bits)-1)/8*8))
//#define RCVLEN(bits) (bits)

void shift(queue &q,PUCHAR buf,unsigned sbit,unsigned bitlen,int read=0){
	WORD w;
	unsigned i,s,j;
	i=sbit/8;
	s=sbit%8;
	bitlen--;
	while(bitlen/8){
		j=bitlen/8;
		if(62<j) j=62;
		q.enq(SERIAL|(read?READ:0)|j&0x3F);
		while(j--){
			w=*(WORD*)(&buf[i++]);
			w>>=s;
			q.enq(w&0xFF);
			bitlen-=8;
		}
	}
	w=*(WORD*)(&buf[i++]);
	w>>=s;
	while(1){
		q.enq((read?READ:0)|OPE|nCS|(bitlen?0:TMS)|(w&1?TDI:0));
		q.enq(OPE|nCS|(w&1?TDI:0)|(bitlen?0:TMS)|TCK);
		w>>=1;
		if(bitlen==0) break;
		bitlen--;
	}
	STATE0(q);	//goto pause state
}
void formatrcvdata(PUCHAR buf,unsigned bitlen){
	unsigned b,s;
	b=(bitlen-1)/8;
//	b=0;
	for(s=b*8;s<bitlen;s++){
		if(buf[b++]&1) bsf(buf,s);
		else bcf(buf,s);
	}
}

static queue que;
static FT_STATUS  ftStatus;
static FT_HANDLE ftHandle;
static unsigned long bc;
static unsigned state;
static HANDLE huo;

#define IDLE 1
#define PIR 2
#define PDR 3

//JTAGP[uI[vAIDLEXe[gɈڍs
int JtagCableOpen(int num){
	huo=UO_Open();
	if(huo!=INVALID_HANDLE_VALUE){
		for(int i=0;i<64;i++) que.enq(0x2F);
		RESET(que);
		RESETtoIDLE(que);
		if(false==UO_Write(huo,que.buf(),que.len(),&bc)){
			UO_Close(huo);
			return false;
		}
		ftHandle=INVALID_HANDLE_VALUE;
		state=IDLE;
		UO_Purge(huo);
	}else{
		try{
			ftStatus = FT_OpenEx("USB-Blaster",FT_OPEN_BY_DESCRIPTION,&ftHandle);
		}catch(...){
			MessageBox(NULL,_T("Could not load FTD2XX.DLL"),NULL,MB_OK);
			return false;
		}
		if(ftStatus != FT_OK){
			return false;
		}
		FT_SetTimeouts(ftHandle,1000,1000);
		que.clear();
		for(int i=0;i<64;i++) que.enq(0x2F);
		RESET(que);
		RESETtoIDLE(que);
		if(FT_OK!=FT_Write(ftHandle,que.buf(),que.len(),&bc)){
			FT_Close(ftHandle);
			return false;
		}
		state=IDLE;
		FT_Purge(ftHandle,FT_PURGE_RX|FT_PURGE_TX);
	}
	que.clear();
	return true;
}

//IRVtgL[
void JtagQueSIR(PUCHAR tx,unsigned bitadr,unsigned bitlen,int read=0){
	switch(state){
		case IDLE : IDLEtoSELDR(que);SELDRtoSIR(que); break;
		case PIR :  PIRtoSELDR(que);SELDRtoSIR(que);  break;
		case PDR :  PDRtoSELDR(que);SELDRtoSIR(que);  break;
		default : throw _T("jtag.cpp : state error!\n");
	}
	shift(que,tx,bitadr,bitlen,read);
	state=PIR;
}

//DRVtgL[
void JtagQueSDR(PUCHAR tx,unsigned bitadr,unsigned bitlen,int read=0){
	switch(state){
		case IDLE : IDLEtoSELDR(que);SELDRtoSDR(que); break;
		case PIR :  PIRtoSELDR(que);SELDRtoSDR(que);  break;
		case PDR :  PDRtoSELDR(que);SELDRtoSDR(que);  break;
		default : throw _T("jtag.cpp : state error!\n");
	}
	shift(que,tx,bitadr,bitlen,read);
	state=PDR;
}

//IDLEL[
void JtagQueIDLE(unsigned clocks){
	switch(state){
		case IDLE : break;
		case PIR :  PIRtoIDLE(que);  break;
		case PDR :  PDRtoIDLE(que);  break;
		default : throw _T("jtag.cpp : state error!\n");
	}
	unsigned j;
	while(clocks/8){
		j=clocks/8;
		if (62<j) j=62;
		que.enq(SERIAL|j&0x3F);
		while(j--){
			que.enq(0xFF);
			clocks-=8;
		}
	}
	while(clocks){
		que.enq(OPE|nCS|TDI);
		que.enq(OPE|nCS|TDI|TCK);
		clocks--;
	}
	state=IDLE;
}

//L[ꂽf[^ۂɑM
int JtagFlushQue(PUCHAR rx=0,unsigned rxbits=0,int check=1){
	DWORD tmp;
	if(huo!=INVALID_HANDLE_VALUE){
		if(false==UO_Write(huo,que.buf(),que.len(),&bc))return false;
	}else{
		if(FT_OK!=FT_Write(ftHandle,que.buf(),que.len(),&bc)) return false;
	}
	if(que.len()!=bc) return false;
	que.clear();
	if(rxbits&&rx){
		if(huo!=INVALID_HANDLE_VALUE){
			if(false==UO_Read(huo,rx,RCVLEN(rxbits),&bc)) return false;
		}else{
			if(FT_OK!=FT_Read(ftHandle,rx,RCVLEN(rxbits),&bc)) return false;
		}
		if(RCVLEN(rxbits)!=bc)	return false;
		formatrcvdata(rx,rxbits);
	}
	if(check){
		if(huo!=INVALID_HANDLE_VALUE){
			bc=0;
		}else{
			FT_GetStatus(ftHandle,&bc,&tmp,&tmp);
		}
		if(bc){
			//throw _T("Remain rx queue! \n");
			return false;
		}
	}
	return true;
}

//JTAGP[uN[Y
void JtagCableClose(){
	if(huo!=INVALID_HANDLE_VALUE){
		UO_Close(huo);
		huo=INVALID_HANDLE_VALUE;
	}else{
		FT_Close(ftHandle);
		ftHandle=INVALID_HANDLE_VALUE;
	}
}
